% substitution.pl
% Subsitutions as lists of V-T pairs where V is a variable and
% T is a term with the meaning V -> T.
% Author: pb

:- module(substitution,[
	subst_self_apply/1,
	subst_apply/3,
	subst_concat/3,
	add_binding/4,
	add_new_binding/4,
	domain/2,
	var_codomain/2,
	dom_member/2,
	dom_member/3,
	im_member/3,
	subst_restricted_to/3
]).

:- use_module(sound_unify).
:- use_module(term).
:- use_module(vars).


subst_self_apply([]).
subst_self_apply([V-T|S]) :-
	sound_unify(V,T),
	subst_self_apply(S).

% applies a substitution to a term
subst_apply(T0,Subst,T1) :-
	var(T0),!,
	(dom_member(Subst,T0,T1) ->
	    true
	;
	    T0 = T1
	).
subst_apply(T0,Subst,T1) :-
	functor(T0,F,A),
	functor(T1,F,A),
	subst_apply(A,T0,Subst,T1).

subst_apply(0,_,_,_) :- !.
subst_apply(A,T0,Subst,T1) :-
	arg(A,T0,ST0),
	arg(A,T1,ST1),
	subst_apply(ST0,Subst,ST1),
	A1 is A-1,
	subst_apply(A1,T0,Subst,T1).


% subst_concat(Ro,Tau,Sigma)
% Sigma = first Ro then Tau
subst_concat(Ro,Tau0,Sigma) :-
	delete_intersection_domain(Tau0,Ro,Tau1),
	subst_concat2(Ro,Tau1,Sigma).

delete_intersection_domain([],_,[]).
delete_intersection_domain([V-_|Tau0],Ro,Tau1) :-
	dom_member(Ro,V),!,
	delete_intersection_domain(Tau0,Ro,Tau1).
delete_intersection_domain([V-T|Tau0],Ro,[V-T|Tau1]) :-
	delete_intersection_domain(Tau0,Ro,Tau1).

subst_concat2([],Tau,Tau).
subst_concat2([V-T0|Ro],Tau,[V-T1|Sigma]) :-
	subst_apply(T0,Tau,T1),
	subst_concat2(Ro,Tau,Sigma).


% add_binding(Var,Term,Subst0,Subst1)
add_binding(Var,Term,Subst0,_Subst1) :-
	dom_member(Subst0,Var,T),!,
	write(Var-T),nl,
	format("add_binding: Var already bound\n",[]),
	write(Var-Term),nl,
	fail.
add_binding(Var,Term,Subst0,Subst1) :-
	add_new_binding(Var,Term,Subst0,Subst1).

add_new_binding(Var,Term0,Subst0,[Var-Term1|Subst1]) :-
	subst_apply(Term0,Subst0,Term1),
	(Term0 == Term1 -> 
	    true
	;
	    write('Peter Graf has lost') % he says for substitution tree insertion this can not happen
	),
	add_new_binding_subst_apply(Subst0,[Var-Term1],Subst1).

add_new_binding_subst_apply([],_,[]).
add_new_binding_subst_apply([V-T0|Subst0],Subst,[V-T1|Subst1]) :-
	subst_apply(T0,Subst,T1),
	(T0 == T1 -> 
	    true
	;
	    write('Peter Graf has lost')
	),
	add_new_binding_subst_apply(Subst0,Subst,Subst1).
	

	
% all variables that are mapped by a subst; i.e. its domain
domain([],[]).
domain([V-_|Subst],[V|Domain]) :-
	domain(Subst,Domain).

% all variables that occur in the codomain of a substitution
var_codomain(Subst,VCOD) :-
	var_codmain(Subst,[],VCOD).

var_codmain([],VCOD,VCOD).
var_codmain([_-T|Subst],VCOD0,VCOD2) :-
	vars_of_term(T,VCOD0,VCOD1),
	var_codmain(Subst,VCOD1,VCOD2).

% is V1 mapped by Subst ?
dom_member([V0-_|Subst],V1) :-
	(V0 == V1 ->
	    true
	;
	    dom_member(Subst,V1)
	).

dom_member([V0-T0|Subst],V1,T1) :-
	(V0 == V1 ->
	    T0 = T1
	;
	    dom_member(Subst,V1,T1)
	).

im_member([V0-T0|Subst],V1,T1) :-
	(T0 == T1 ->
	    V0 = V1
	;
	    im_member(Subst,V1,T1)
	).

	
subst_restricted_to([],_Vars,[]).
subst_restricted_to([V-T|Subst0],Vars,Subst1) :-
	var_memberchk(V,Vars),!,
	[V-T|Subst2] = Subst1,
	subst_restricted_to(Subst0,Vars,Subst2).
subst_restricted_to([_|Subst0],Vars,Subst1) :-
	subst_restricted_to(Subst0,Vars,Subst1).
 
